from collections import deque

from geometry import distance


def calculate_bezier(p, steps=30):
    """calculate a bezier curve from 4 control points

    Returns a list of the resulting points.

    The function uses the forward differencing algorithm described here:
    http://www.niksula.cs.hut.fi/~hkankaan/Homepages/bezierfast.html

    This code taken from www.pygame.org/wiki/BezierCurve.
    """
    #
    t = 1.0 / steps
    temp = t*t
    #
    f = p[0]
    fd = 3 * (p[1] - p[0]) * t
    fdd_per_2 = 3 * (p[0] - 2 * p[1] + p[2]) * temp
    fddd_per_2 = 3 * (3 * (p[1] - p[2]) + p[3] - p[0]) * temp * t
    #
    fddd = fddd_per_2 + fddd_per_2
    fdd = fdd_per_2 + fdd_per_2
    fddd_per_6 = fddd_per_2 * (1.0 / 3)
    #
    for x in range(steps):
        yield f
        f = f + fd + fdd_per_2 + fddd_per_6
        fd = fd + fdd + fddd_per_2
        fdd = fdd + fddd
        fdd_per_2 = fdd_per_2 + fddd_per_2
    yield f


def plot_curve(p, n=200):
    """plot a curved path along one or more sets of control points

    p is a one-dimensional array of points in multiples of four: i.e.
    four points per curve.

    steps is the number of points desired. steps is weighted by the sum
    distance of each four points in order to yield approximately
    equidistant points.

    This code derived from www.pygame.org/wiki/BezierCurve.
    """
    plot = deque()
    for x in range((len(p)-1)/3):
        slice_ = p[3*x:3*x+4]
        dist = 0.0
        p0 = slice_[0]
        for i in range(1, 3):
            dist = max(dist, distance(p0, p[i]))
            p0 = p[i]
        b_points = calculate_bezier(slice_, n)
        plot.append(b_points)
    return plot
