import collections
from os.path import join as join_path
import sys
import random
from random import randrange

import gsdl2
import settings
from gamesound import jukebox
from geometry import distance, angle_of, point_on_circumference, circle_collided_other
from bezier import plot_curve
from plotutil import calc_line
import sprite
import text
import textureutil
import tweening
from vec2d import Vec2d


tweener = tweening.Tweener()
laser_shots = []
explosions = []


def update(dt):
    tweener.update(dt)
    for shot in laser_shots:
        shot.update(dt)
    for exp in explosions:
        exp.update(dt)


def clear():
    del laser_shots[:]
    del explosions[:]
    tweener.clear()


def iround(n):
    return int(round(n))


def w(*args):
    s = ' '.join(args)
    sys.stdout.write(s)
    sys.stdout.flush()


class _StateBase(object):

    @staticmethod
    def enter(self):
        pass

    @staticmethod
    def exit(self):
        pass

    @staticmethod
    def think(self, dt):
        new_state = self.start_state
        if new_state:
            self.start_state = None
            return new_state


class _StateConvoy(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateConvoy({})'.format(self))

    @staticmethod
    def think(self, dt):
        self.shoot()
        return self.move(dt, _StateGotoStandby)

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateConvoy({})'.format(self))


class _StateGotoStandby(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateGotoStandby({})'.format(self))
        ranky, rankx = self.rank[0], self.rank[1]
        x, y = settings.standby_pos[ranky][rankx]
        self.move_to = x, y

    @staticmethod
    def think(self, dt):
        # adjust for shifting formation
        ranky, rankx = self.rank
        self.move_to = settings.standby_pos[ranky][rankx]
        # calculate new position and angle
        self.trail.append(self.rect.center)
        fm = self.old_pos
        to = self.move_to
        radius = settings.enemy_speed * dt
        d = distance(fm, to)
        if radius > d:
            self.pos[:] = to
        else:
            angle = angle_of(fm, to)
            self.pos[:] = point_on_circumference(fm, radius, angle)
        self.old_pos[:] = self.pos
        self.rect.center = self.pos
        self.angle = angle_of(self.trail[0], self.pos)
        if self.pos == self.move_to:
            return _StateStandby

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateGotoStandby({})'.format(self))


class _StateStandby(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateStandby({})'.format(self))
        self.reset_pos(self.pos)
        self.leg = 0

    @staticmethod
    def think(self, dt):
        self.turn_towards(dt, 0)
        ranky, rankx = self.rank
        self.old_pos = self.pos
        self.rect.center = self.move_to = self.pos = settings.standby_pos[ranky][rankx]

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateStandby({})'.format(self))


class _StatePatrol(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StatePatrol({})'.format(self))
        self.new_path()
        self.shots = 0

    @staticmethod
    def think(self, dt):
        new_state = self.move(dt, _StateEndPatrol)
        if new_state is None:
            self.shoot()
        elif isinstance(self, Kamikaze):
            return _StatePark
        return new_state

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StatePatrol({})'.format(self))


class _StatePark(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StatePark({})'.format(self))

    @staticmethod
    def think(self, dt):
        new_state = self.move(dt, _StateEndPatrol)
        if new_state is _StateEndPatrol:
            new_state = self.new_path()
            if new_state is not None:
                return new_state
        self.shoot()

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StatePark({})'.format(self))


class _StateRam(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateRam({})'.format(self))

    @staticmethod
    def think(self, dt):
        new_state = self.move(dt, _StateEndPatrol)
        if new_state is _StateEndPatrol:
            new_state = self.new_path()
            if new_state is not None:
                return new_state
        self.shoot()

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateRam({})'.format(self))


class _StateEndPatrol(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateEndPatrol({})'.format(self))
        self.reset_pos((self.pos.x, -50))
        fm = Vec2d(self.pos.x, -50)
        to = Vec2d(self.pos.x, 0)
        points = [Vec2d(fm), Vec2d(to)]
        fm = to
        ranky, rankx = self.rank
        # TODO: Vec2d not needed?
        # to = Vec2d(settings.standby_pos[ranky][rankx])
        to = settings.standby_pos[ranky][rankx]
        for i in range(2, 4):
            midx = fm.x + int((to.x - fm.x) * i/3)
            midy = fm.y + int((to.y - fm.y) * i/3)
            points.append(Vec2d(midx, midy))
        self.new_path(points)

    @staticmethod
    def think(self, dt):
        return _StateReform

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateEndPatrol({})'.format(self))


class _StateReform(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_enemy_states:
            print('Enemy: enter _StateReform({})'.format(self))

    @staticmethod
    def think(self, dt):
        return self.move(dt, _StateGotoStandby)

    @staticmethod
    def exit(self):
        if settings.print_enemy_states:
            print('Enemy: exit _StateReform({})'.format(self))


class Projectile(sprite.TextureSprite):
    layer = settings.layer_projectiles
    collided = staticmethod(circle_collided_other)
    _texture = None
    radius = None
    speed = None
    turn_speed = None
    alive = True

    def __init__(self, ship, angle, radius=-1):
        jukebox['enemy_laser'].play()
        self.ship = ship
        sprite.TextureSprite.__init__(self, self.texture, ship.pos, angle, radius=radius)

    def _get_origin(self):
        return self.pos
    origin = property(_get_origin)

    def hit(self, others):
        for other in others:
            if isinstance(other, settings.FriendlyShipTypes):
                # print('enemy.LaserBlast: hit', other)
                other.kill(True)
                self.kill(True)
            elif isinstance(other, settings.FriendlyFireTypes) and isinstance(self, settings.EnemyOrdinanceTypes):
                other.kill(True)
                self.kill(True)

    def kill(self, hit=False, score=True):
        if self.alive:
            self.alive = False
            laser_shots.remove(self)
            settings.space.remove(self)
            if hit and isinstance(self, settings.EnemyOrdinanceTypes):
                EnemyExplodes(self.pos)
                if score:
                    score = scores[self.__class__]
                    settings.score += score
                    text.make_floating_score(str(score), self.rect.midtop)


class Missile(Projectile):
    _texture = None

    def __init__(self, ship, angle=None):
        if Missile._texture is None:
            surf = gsdl2.image.load(settings.img_projectiles)
            surf.set_colorkey(gsdl2.Color('black'))
            tex_projectiles = gsdl2.texture.Texture(gsdl2.display.get_renderer(), surf)
            Missile._texture = textureutil.copy(tex_projectiles, gsdl2.Rect(211, 48, 20, 20), scale=1.5)
        angle = ship.angle % 360 if angle is None else angle % 360
        if angle == 0:
            angle = 360
        Projectile.__init__(self, ship, angle, radius=10)
        self.speed = settings.player_speed * 0.65
        self.turn_speed = settings.player_turn_rate * 0.5
        self.alive = True
        laser_shots.append(self)
        settings.space.add(self)

    def update(self, dt):
        self.old_pos = self.pos
        self._turn_towards(settings.player_ship.pos, dt)
        self.pos = point_on_circumference(self.pos, self.speed * dt, self.angle)
        self.rect.center = self.move_to = self.pos
        settings.space.add(self)

    def _turn_towards(self, pos, dt):
        angle = angle_of(self.pos, pos)
        diff = self.angle - angle
        if diff != 0:
            if diff > 0:
                if diff < 180:
                    adjust = -self.turn_speed * dt
                else:
                    adjust = self.turn_speed * dt
            else:
                if diff > -180:
                    adjust = self.turn_speed * dt
                else:
                    adjust = -self.turn_speed * dt
            degrees = self.angle + adjust
            degrees %= 360
            self.angle = degrees


class Mine(Projectile):
    _texture = None

    def __init__(self, ship, angle):
        if Mine._texture is None:
            surf = gsdl2.image.load(settings.img_mine)
            surf.set_colorkey(gsdl2.Color('black'))
            Mine._texture = gsdl2.texture.Texture(gsdl2.display.get_renderer(), surf)
        Projectile.__init__(self, ship, angle)
        self.speed = settings.player_speed / 3.0
        self.alive = True
        laser_shots.append(self)
        settings.space.add(self)

    def update(self, dt):
        self.old_pos = self.pos
        self.pos = point_on_circumference(self.pos, self.speed * dt, self.angle)
        self.rect.center = self.move_to = self.pos
        settings.space.add(self)
        if not settings.screen_rect.collidepoint(self.rect.center):
            self.kill()


class LaserBlast(Projectile):
    _texture = None
    layer = settings.layer_lasers

    def __init__(self, ship, angle):
        self.speed = settings.laser_shot_speed
        if LaserBlast._texture is None:
            LaserBlast._texture = gsdl2.texture.Texture(
                gsdl2.display.get_renderer(), gsdl2.image.load(settings.img_laser_orange))
        Projectile.__init__(self, ship, ship.angle)
        self.alive = True
        laser_shots.append(self)
        settings.space.add(self)

    def update(self, dt):
        self.old_pos = self.pos
        self.pos = point_on_circumference(self.pos, self.speed * dt, self.angle)
        self.rect.center = self.move_to = self.pos
        if not settings.screen_rect.collidepoint(self.rect.center):
            self.kill()
        else:
            settings.space.add(self)


class MissileExplodes(object):
    def __init__(self, pos):
        jukebox['enemy explodes'].play()
        self.pos = pos
        self.images = SpriteStripAnim(join_path('image', 'Explode2x200.png'),
            (1,1,24,24), 8, -1, False, game.fps.fps/12, name='missile_explodes')
        self.images.iter()

    def update(self, *args):
        try:
            self.image = self.images.next()
        except StopIteration:
            self.kill()
        self.rect = self.image.get_rect()
        self.rect.center = self.pos


class EnemyExplodes(sprite.TextureSprite):
    layer = settings.layer_explosions
    textures = []

    def __init__(self, pos):
        jukebox['enemy_hit'].play()
        if not EnemyExplodes.textures:
            sheet = gsdl2.image.load(settings.img_enemy_explode)
            sheet.set_colorkey((0, 0, 0))
            sheet = gsdl2.Texture(gsdl2.display.get_renderer(), sheet)
            EnemyExplodes.textures[:] = textureutil.copy_strip(sheet, gsdl2.Rect(0, 1, 48, 48), numx=8, scale=1.75)
        sprite.TextureSprite.__init__(self, self.textures[0], pos, alpha=32*7)

        self.tween_val = 0
        self.tween = tweener.create_tween(self, 'tween_val', 0, 8, 0.5, cb_end=self.remove)

        self.alive = False
        explosions.append(self)
        settings.space.add(self)

    def hit(self, other):
        pass

    def update(self, dt):
        try:
            self.texture = self.textures[int(self.tween_val)]
        except IndexError:
            pass

    def remove(self, *args):
        self.alive = False
        try:
            explosions.remove(self)
        except ValueError as e:
            pass
        settings.space.remove(self)


class Enemy(sprite.TextureSprite):
    """base class for enemy ships"""
    _texture = None
    layer = settings.layer_enemies
    collided = staticmethod(circle_collided_other)

    def __init__(self, rank, pos):
        sprite.TextureSprite.__init__(self, self._texture, pos)
        self.rank = rank

        self.points = None  # control points for Bezier curves
        self.path = None    # deque of plotted point generators
        self.steps = None   # the current generator from path
        self.speed = settings.enemy_speed
        self.throttle = 1.0  # local speed factor

        self.shots = 0
        self.fire_rate = round(settings.enemy_fire_rate)
        self.spec_rate = round(settings.enemy_fire_rate / 2.0)

        self.turn_rate = 180.0  # rate of turn: degrees per second
        self.trail = collections.deque(maxlen=5)
        self.trail.append(self.rect.center)

        self.countdown = settings.max_fps
        self.current_state = _StateBase
        self.start_state = _StateConvoy

        self.alive = True

    def reset_pos(self, pos):
        self.pos = self.old_pos = self.move_to = self.rect.center = pos
        self.trail.clear()
        self.trail.append(pos)
        del self.points[:]
        self.path.clear()
        self.steps = None

    def _get_origin(self):
        return self.pos
    origin = property(_get_origin)

    def is_patrol(self):
        return self.current_state is _StatePatrol

    def is_standby(self):
        return self.current_state is _StateStandby

    def is_convoy(self):
        return self.current_state is _StateConvoy

    def is_reform(self):
        return self.current_state is _StateReform

    def is_ram(self):
        return self.current_state is _StateRam

    def go_patrol(self):
        if self.is_standby():
            self.set_state(_StatePatrol)

    def update(self, dt):
        state = self.current_state.think(self, dt)
        if state is not None:
            self.set_state(state)
        self.rect.center = self.pos
        if self.alive:
            settings.space.add(self)

    def move(self, dt, next_state):
        fm = self.old_pos
        to = self.move_to
        radius = self.speed * self.throttle * dt
        d = distance(fm, to)
        try:
            while radius > d:
                # get new step
                # print('move: get new step')
                to = next(self.steps)
                self.move_to = to
                d = distance(fm, to)
        except StopIteration:
            try:
                # get new steps generator
                self.steps = self.path.popleft()
                to = next(self.steps)
                self.move_to = to
                d = distance(fm, to)
            except IndexError:
                # end of path
                return next_state
        self.trail.append(self.rect.center)
        angle = angle_of(fm, to)
        self.pos[:] = point_on_circumference(fm, d, angle)
        self.old_pos[:] = self.pos
        self.rect.center = self.pos
        if self.trail[0] != self.pos:
            self.angle = angle_of(self.trail[0], self.pos)

    def set_state(self, new_state):
        if self.current_state:
            self.current_state.exit(self)
        self.current_state = new_state
        new_state.enter(self)

    def new_path(self, points=None, path=None):
        if points is None:
            self.points = []
        else:
            self.points = points
        if path is None:
            self.path = plot_curve(self.points)
        else:
            self.path = path
        # print('Enemy.new_path: points', self.points)
        # print('Enemy.new_path: path', self.path)
        # traceback.print_stack()
        if self.path:
            self.steps = self.path.popleft()
            self.old_pos = Vec2d(next(self.steps))
            self.move_to = Vec2d(next(self.steps))

    def turn_towards(self, dt, angle):
        angle %= 360.0
        if angle != self.angle:
            diff = self.angle - angle
            diff %= 360.0
            if 0 < diff < 180:
                turn = -self.turn_rate * dt
            else:  # if 180 <= diff < 360:
                turn = self.turn_rate * dt
            if abs(diff) < abs(turn):
                self.angle = angle
            else:
                self.angle += turn
                self.angle %= 360.0

    def hit(self, others):
        for other in others:
            if other in settings.FriendlyTypes:
                # print('enemy.Enemy: hit', other)
                other.kill(True)
                self.kill(True)
                # if issubclass(type(other), (player.LaserBlast,player.Ship)):
                #     game.score.add(_SCORES[self.state], self._bonus(), self.rect.center)
                #     game.groups.add(EnemyExplodes(
                #         self.rect.center), layer='enemies', under=True)
                # if randrange(5) == 0:
                #     ExtraShipDebris(self.image, self.rect.center)

    def kill(self, hit=False):
        if self.alive:
            self.alive = False
            settings.space.remove(self)
            settings.enemy_fleet.remove(self)
            if hit:
                EnemyExplodes(self.pos)
                score = scores[self.current_state]
                settings.score += score
                text.make_floating_score(str(score), self.rect.midtop)

    def _bonus(self):
        """return bonus multiplier; override to tweak awarded points"""
        return 1

    def power_up(self, others):
        for other in others:
            other.hit(self)

    def shoot(self):
        """enemy attack logic; override and call the parent method"""
        if self.fire_rate > 0:
            self.fire_rate -= 1
        if self.spec_rate > 0:
            self.spec_rate -= 1

    def _strafing_attack(self):
        """aimless strafing attack"""
        # path_len = len(self.path)
        # step = int(round(self.step))
        # if path_len < 10:
        #     x1, y1 = self.path[0]
        #     x2, y2 = self.path[-1]
        # elif path_len > step + 9:
        #     x1, y1 = self.path[step]
        #     x2, y2 = self.path[step+9]
        # else:
        #     x1, y1 = self.path[-1]
        #     x2, y2 = self.path[-10]
        # if y1 < y2:
        LaserBlast(self, 0)
        self.shots += 1
        self.fire_rate = round(settings.enemy_fire_rate)

    def __str__(self, *args, **kwargs):
        x, y = self.pos
        return '<{}({}, {})> [{}]'.format(
            self.__class__.__name__, self.rank, Vec2d(int(x), int(y)), gsdl2.time.get_ticks())


class Fighter(Enemy):
    _texture = None

    def __init__(self, rank, pos):
        if Fighter._texture is None:
            sheet = gsdl2.image.load_texture(settings.img_player)
            w, h = 256, 256
            Fighter._texture = textureutil.copy(sheet, gsdl2.Rect(0, h * 0, w, h).scale(0.5, 0.5), scale=0.38)
        Enemy.__init__(self, rank, pos)
        self.leg = 0

    def new_path(self, points=None, path=None):
        game_left = settings.screen_rect.left
        game_right = settings.screen_rect.right

        def make_points():
            cx, cy = center = self.rect.center
            game_centerx = settings.screen_rect.centerx
            # Default path: enemy fighter flies an eccentric ? pattern
            points = [
                Vec2d(center),
                Vec2d(443, 0), Vec2d(527, 0), Vec2d(527, 175),
                Vec2d(527, 266), Vec2d(428, 369), Vec2d(286, 452),
                Vec2d(164, 521), Vec2d(86, 625), Vec2d(86, settings.screen.h + self.rect.h),
            ]
            # Rules to tweak runtime path.
            #
            # Enemy points start out right-facing. If enemy is left-facing,
            # swap the x values to the other side. Then introduce some
            # randomness to the pattern as follows:
            #
            # tweak 2.x near to sprite.x (subject to left/right-facedness)
            # tweak 9.x to ram player
            # replace 1.x with sprite.x (subject to left/right-facedness)
            # set 3.x and 4.x to 2.x
            # vary 5.x through 9.x to kick out ? tail
            # vary 3.y through 7.y for height of ? tail
            if cx >= game_centerx:
                # Default path already facing the right way; set 2.x, 9.x
                points[2].x = cx + randrange(40, 80)
                if randrange(5) == 0:
                    points[9].x = settings.player_spawn[0]
                else:
                    points[9].x = game_left
            else:
                # Flip default path to face other way; set 2.x, 9.x
                for p in points[2:]:
                    p.x = game_centerx - (p.x - game_centerx)
                points[2].x = cx - randrange(40, 80)
                if randrange(5) == 0:
                    points[9].x = settings.player_spawn[1]
                else:
                    points[9].x = game_right
            # Tweak the rest to add some unpredictability
            points[1].x = cx
            points[3].x = points[4].x = points[2].x
            randx = randrange(0, 60) - 30
            for i in (5, 6, 7, 8, 9):
                points[i].x += randx
            randy = randrange(0, 120) - 100
            for i in (3, 4, 5, 6, 7):
                points[i].y += randy
            return points

        if points is None:
            if self.leg == 0:
                points = make_points()
                self.leg = 1
            elif self.leg == 1:
                self.set_state(_StateEndPatrol)
                self.leg = 0
        # Knock x inbounds
        if points is not None:
            for p in points[1:]:
                if p.x > game_right:
                    p.x = game_right
                if p.x < game_left:
                    p.x = game_left
        Enemy.new_path(self, points, path)

    def shoot(self):
        Enemy.shoot(self)
        if self.is_patrol():
            mx = 0.4
        else:
            mx = 0.9
        fire_now = self.shots < settings.enemy_max_shots and randrange(round(1.0 / settings.time_step * mx)) == 0
        if fire_now:
            angle_to_player = round(angle_of(self.rect.center, settings.player_spawn))
            spec_now = self.shots < settings.enemy_max_shots and \
                       angle_to_player in range(int(self.angle) - 5, int(self.angle) + 5)
            if spec_now and self.spec_rate <= 0:
                # Special aimed attack
                LaserBlast(self, 0)
                self.shots += 1
                self.spec_rate = round(settings.enemy_fire_rate / 2)
            elif self.fire_rate <= 0:
                # Normal attack
                self._strafing_attack()

    def _bonus(self):
        """double default score if on patrol"""
        if self.state == PATROL:
            bonus = 2
        else:
            bonus = 1
        return bonus


class Sweeper(Enemy):
    _texture = None

    def __init__(self, rank, pos):
        if Sweeper._texture is None:
            sheet = gsdl2.image.load_texture(settings.img_player)
            w, h = 256, 256
            Sweeper._texture = textureutil.copy(sheet, gsdl2.Rect(0, h * 8, w, h).scale(0.5, 0.5), scale=0.35)
        Enemy.__init__(self, rank, pos)
        self.leg = 0

    def new_path(self, points=None, path=None):
        screen_rect = settings.screen_rect
        game_left = screen_rect.left
        game_right = screen_rect.right

        def make_points():
            cx, cy = center = self.rect.center
            game_centerx = screen_rect.centerx
            # Default path: enemy sweeper flies a wide S pattern
            points = [
                Vec2d(center),
                Vec2d(400, 0), Vec2d(527, 0), Vec2d(527, 76),
                Vec2d(527, 241), Vec2d(383, 241), Vec2d(193, 284),
                Vec2d(17, 325), Vec2d(17, 446), Vec2d(129, 509),
                Vec2d(275, 573), Vec2d(574, 573), Vec2d(574, settings.screen.h + self.rect.h),
            ]
            # Rules to tweak runtime path.
            #
            # Enemy points start out right-facing. If enemy is left-facing,
            # swap the x values to the other side. Then introduce some
            # randomness to the pattern as follows:
            #
            # tweak 2.x near to sprite.x (subject to left/right-facedness)
            # tweak 12.x to ram player
            # replace 1.x with sprite.x (subject to left/right-facedness)
            # set 3.x and 4.x to 2.x
            # vary 6.x through 9.x for width of S tail
            # vary 5.y through 7.y for top of S tail
            # vary 8.y through 11.y for bottom of S tail
            if cx >= game_centerx:
                # Default path already facing the right way; set 2.x, 9.x
                points[2].x = cx + randrange(20, 60)
                if randrange(5) == 0:
                    points[12].x = settings.player_ship.rect.centerx
                else:
                    points[12].x = game_left
            else:
                # Flip default path to face other way; set 2.x, 9.x
                for p in points[2:]:
                    p.x = game_centerx - (p.x - game_centerx)
                points[2].x = cx - randrange(20, 60)
                if randrange(5) == 0:
                    points[12].x = settings.player_ship.rect.centerx
                else:
                    points[12].x = game_right
            # Tweak the rest to add some unpredictability
            points[1].x = cx
            points[3].x = points[4].x = points[2].x
            randx = randrange(0, 60) - 30
            for i in (6, 7, 8, 9):
                points[i].x += randx
            randy = randrange(0, 80) - 60
            for i in (5, 6, 7):
                points[i].y += randy
            randy = randrange(0, 80) - 40
            for i in (8, 9, 10, 11):
                points[i].y += randy
            return points
        if points is None:
            if self.leg == 0:
                points = make_points()
                self.leg = 1
            elif self.leg == 1:
                self.set_state(_StateEndPatrol)
                self.leg = 0
        # Knock x inbounds
        if path is not None:
            for p in points[1:]:
                if p.x > game_right:
                    p.x = game_right
                if p.x < game_left:
                    p.x = game_left
        Enemy.new_path(self, points, path)

    def shoot(self):
        Enemy.shoot(self)
        if self.is_patrol():
            mx = 0.4
        else:
            mx = 1.8  # 1.3
        fire_now = self.shots < settings.enemy_max_shots and randrange(round(settings.max_fps * mx)) == 0
        if fire_now:
            angle_to_player = angle_of(self.pos, settings.player_ship.pos) + 360.0
            self_angle = self.angle + 360.0
            spec_now = self.shots < settings.enemy_max_shots and \
                       self_angle - 5 <= angle_to_player <= self_angle + 5
            if spec_now and self.spec_rate <= 0:
                # Special aimed attack
                LaserBlast(self, 0)
                self.shots += 1
                self.spec_rate = round(settings.enemy_fire_rate * 1.0 / settings.time_step / 2)
            elif self.fire_rate <= 0:
                # Normal attack
                self._strafing_attack()


class Kamikaze(Enemy):
    _texture = None

    def __init__(self, rank, pos):
        if Kamikaze._texture is None:
            sheet = gsdl2.image.load_texture(settings.img_player)
            w, h = 256, 256
            Kamikaze._texture = textureutil.copy(sheet, gsdl2.Rect(0, h * 4, w, h).scale(0.66, 0.66), scale=0.3)
        Enemy.__init__(self, rank, pos)
        self.leg = 0

    def new_path(self, points=None, path=None):
        screen_rect = settings.screen_rect
        game_left = screen_rect.left
        game_right = screen_rect.right
        if points is None:
            cx, cy = center = self.rect.center
            game_centerx = screen_rect.centerx
            if self.leg == 0:
                # Go to park
                if cx > game_centerx:
                    points = [
                        Vec2d(center), Vec2d(cx, cy - 100),
                        Vec2d(cx + 100, cy - 100), Vec2d(cx + 100, settings.standby_y[-1] + 25)
                    ]
                else:
                    points = [
                        Vec2d(center), Vec2d(cx, cy - 100),
                        Vec2d(cx - 100, cy - 100), Vec2d(cx - 100, settings.standby_y[-1] + 25)
                    ]
                self.leg += 1
            elif self.leg == 1:
                # Park
                self.pause = 0.5
                self.leg += 1
                missile_count = 0
                launch_total = int(settings.max_on_patrol[type(self)])
                angles = random.sample(range(0, 360, 20), launch_total)
                while missile_count < launch_total:
                    Missile(self, angles[missile_count])
                    missile_count += 1
            elif self.leg == 2:
                # Vulture
                if self.pause > 0:
                    self.turn_towards(settings.time_step, angle_of(self.pos, settings.player_ship.pos))
                    self.pause -= 1
                else:
                    self.leg += 1
            elif self.leg == 3:
                # Ram
                srect = settings.player_ship.rect
                self.trail.clear()
                self.trail.append(center)
                if randrange(0, 2):
                    path = collections.deque([iter(calc_line(center, srect.center))])
                    self.throttle = 1.5
                else:
                    dy = srect.centery - cy
                    if cx > game_centerx:
                        points = [
                            Vec2d(center), Vec2d(cx + 200, cy + dy * 1/3),
                            Vec2d(cx + 200, cy + dy * 2/3), Vec2d(srect.center),
                        ]
                    else:
                        points = [
                            Vec2d(center), Vec2d(cx - 200, cy + dy * 1/3),
                            Vec2d(cx - 200, cy + dy * 2/3), Vec2d(srect.center),
                        ]
                    self.throttle = 2.0
                self.leg += 1
                self.set_state(_StateRam)
            elif self.leg == 4:
                # Escape
                to_rect = settings.screen_rect
                path = collections.deque([iter(calc_line(self.rect.center, (cx, to_rect.height + self.rect.h)))])
                self.leg += 1
            elif self.leg == 5:
                # Reform
                self.set_state(_StateEndPatrol)
                self.throttle = 1.0
                self.leg = 0
        # Knock x inbounds
        if points is not None:
            for p in points[1:]:
                if p.x > game_right:
                    p.x = game_right
                if p.x < game_left:
                    p.x = game_left
        Enemy.new_path(self, points, path)

    def shoot(self):
        if self.is_convoy():
            if not randrange(1.0 / settings.time_step * 30) == 0:
                return
            missile_count = 0
            launch_total = int(settings.max_on_patrol[type(self)])
            angles = random.sample(range(0, 360, 20), launch_total)
            while missile_count < launch_total:
                Missile(self, angles[missile_count])
                missile_count += 1
        else:
            Enemy.shoot(self)

    def _bonus(self):
        """3X default score if on patrol"""
        if self.is_patrol() or self.is_ram():
            bonus = 3
        else:
            bonus = 1
        return bonus


class Destroyer(Enemy):
    _texture = None

    def __init__(self, rank, pos):
        if Destroyer._texture is None:
            sheet = gsdl2.image.load_texture(settings.img_player)
            w, h = 256, 256
            Destroyer._texture = textureutil.copy(sheet, gsdl2.Rect(0, h * 10, w, h).scale(0.85, 0.85), scale=0.28)
        Enemy.__init__(self, rank, pos)
        self.leg = 0
        self.spec_reset = 1.0 / settings.time_step * 5
        self.spec_rate = self.spec_reset
        self.leg = 0
        self.shoot()

    def new_path(self, points=None, path=None):
        screen_rect = settings.screen_rect
        game_left = screen_rect.left
        game_right = screen_rect.right

        def make_points():
            cx, cy = center = self.rect.center
            game_centerx = screen_rect.centerx
            points = [
                Vec2d(center),
                Vec2d(555, 0), Vec2d(682, 0), Vec2d(682, 71),
                Vec2d(682, 139), Vec2d(614, 209), Vec2d(553, 272),
                Vec2d(485, 342), Vec2d(427, 407), Vec2d(489, 487),
                Vec2d(555, 581), Vec2d(617, 660), Vec2d(709, settings.screen.h + self.rect.h)
            ]
            if cx < game_centerx:
                for p in points[2:]:
                    p.x = game_centerx - (p.x - game_centerx)
            return points
        if points is None:
            if self.leg == 0:
                points = make_points()
                self.leg = 1
            elif self.leg == 1:
                self.set_state(_StateEndPatrol)
                self.leg = 0
        # Knock x inbounds
        if points is not None:
            for p in points[1:]:
                if p.x > game_right:
                    p.x = game_right
                if p.x < game_left:
                    p.x = game_left
        Enemy.new_path(self, points, path)

    def shoot(self):
        Enemy.shoot(self)
        fire_now = self.spec_rate <= 0
        if fire_now:
            angle = angle_of(self.pos, settings.player_ship.pos)
            # Every five waves Destroyer drops an extra mine
            # mine_count = 0
            launch_total = int(settings.enemy_fleet.max_on_patrol[type(self)])
            variances = random.sample(range(0, 20, 2), launch_total)
            # while mine_count < launch_total:
            for mine_count in range(launch_total):
                variance = variances[mine_count] - 10
                Mine(self, angle + variance)
                # mine_count += 1
            self.spec_rate = self.spec_reset
        else:
            self.spec_rate -= 1

    def _bonus(self):
        """4X default score if on patrol"""
        if self.is_patrol():
            bonus = 4
        else:
            bonus = 1
        return bonus


# Initialize classifiers in settings module
settings.EnemyShipTypes = (
    Enemy,
)
settings.EnemyEnergyTypes = (
    LaserBlast,
)
settings.EnemyOrdinanceTypes = (
    Missile,
    Mine,
)
settings.EnemyTypes = settings.EnemyShipTypes + settings.EnemyEnergyTypes + settings.EnemyOrdinanceTypes


scores = {
    _StateBase: 0,
    _StateStandby: 1,
    _StateReform: 1,
    _StateConvoy: 50,
    _StateGotoStandby: 50,
    _StateEndPatrol: 50,
    _StatePatrol: 100,
    _StatePark: 500,
    _StateRam: 500,
    Missile: 500,
    Mine: 500,
}
