"""game.py - a context for running a game session or level
"""

import collections
import sys

import gsdl2
from gsdl2.locals import S_ESCAPE, S_SPACE, S_K, S_G, S_P

import context
import player
import enemy
import enemyfleet
from gamesound import jukebox
import settings
import sky
import spatialhash
import sprite
import text
import tweening


tweener = tweening.Tweener()


class _StateBase(object):
    @staticmethod
    def enter(self):
        pass

    @staticmethod
    def exit(self):
        pass

    @staticmethod
    def think(self, dt):
        new_state = self.start_state
        if new_state:
            self.start_state = None
            return new_state


class _StateIntro(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_game_states:
            print('Game: enter _StateIntro({})'.format(self))
        tweener.create_tween(self, 'tween_intro', 0.0, 1.0, settings.intro_duration)
        tex = gsdl2.Texture(self.renderer, settings.font_stage.render(
            'Wave {}'.format(settings.stage + 1), settings.color_stage))
        self.tex_stage = sprite.TextureSprite(tex, settings.screen_rect.center)
        settings.space.add(self.tex_stage)
        settings.player_ship.reset()
        jukebox['new_stage'].play()

    @staticmethod
    def think(self, dt):
        self.toss_events()
        if self.tween_intro == 1.0:
            return _StateBattle

    @staticmethod
    def exit(self):
        if settings.print_game_states:
            print('Game: exit _StateIntro({})'.format(self))
        settings.space.remove(self.tex_stage)


class _StateBattle(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_game_states:
            print('Game: enter _StateBattle({})'.format(self))

    @staticmethod
    def think(self, dt):
        self.get_events()
        settings.player_ship.mouse_motion(self.mouse_pos)
        player.update(dt)
        enemyfleet.update(dt)
        enemy.update(dt)
        sprite.update(dt)
        text.update(dt)
        self.update_collisions()
        return self.update_wave()

    @staticmethod
    def exit(self):
        if settings.print_game_states:
            print('Game: exit _StateBattle({})'.format(self))


class _StateEndWave(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_game_states:
            print('Game: enter _StateEndWave({})'.format(self))
        settings.enemy_fleet.quiesce = True

    @staticmethod
    def think(self, dt):
        self.toss_events()
        player.update(dt)
        # enemyfleet.update(dt)
        enemy.update(dt)
        sprite.update(dt)
        text.update(dt)
        # self.update_collisions()
        if not any((player.explosions, enemy.explosions)):
            return _StateIntro

    @staticmethod
    def exit(self):
        if settings.print_game_states:
            print('Game: exit _StateEndWave({})'.format(self))
        sprite.clear()
        enemy.clear()
        text.clear()
        enemyfleet.clear()
        settings.space.clear()
        settings.player_ship.update_pos(settings.player_spawn)
        settings.space.add(settings.player_ship)
        settings.enemy_fleet.next_wave()
        settings.enemy_fleet.quiesce = False


class _StateRespawn(_StateBase):
    @staticmethod
    def enter(self):
        if settings.print_game_states:
            print('Game: enter _StateRespawn({})'.format(self))
        settings.enemy_fleet.quiesce = True

    @staticmethod
    def think(self, dt):
        self.toss_events()
        player.update(dt)
        enemyfleet.update(dt)
        enemy.update(dt)
        sprite.update(dt)
        text.update(dt)
        self.update_collisions()
        if all([e.is_standby() for e in settings.enemy_fleet.ships_deployed]) and \
                not enemy.explosions and not player.explosions:
            return _StateBattle

    @staticmethod
    def exit(self):
        if settings.print_game_states:
            print('Game: exit _StateRespawn({})'.format(self))
        for shot in list(enemy.laser_shots):
            shot.kill(hit=True, score=False)
        settings.player_ship = player.Ship()
        settings.space.add(settings.player_ship)
        settings.enemy_fleet.quiesce = False


def get_layer(obj):
    return obj.layer


def get_circle_overlay(obj):
    if obj.radius in circle_overlays:
        answer = circle_overlays[obj.radius]
    else:
        overlay_rect = gsdl2.Rect(0, 0, obj.radius * 2, obj.radius * 2)
        surf = gsdl2.Surface(overlay_rect.size)
        gsdl2.draw.circle(surf, gsdl2.Color('red'), overlay_rect.center, obj.radius)
        surf.set_colorkey((0, 0, 0))
        overlay = gsdl2.texture.Texture(gsdl2.display.get_renderer(), surf)
        answer = overlay, overlay_rect
        circle_overlays[obj.radius] = answer
    return answer
circle_overlays = {}


def get_rect_overlay(obj):
    if obj.radius in rect_overlays:
        answer = rect_overlays[obj.radius]
    else:
        overlay_rect = gsdl2.Rect(0, 0, obj.radius * 2, obj.radius * 2)
        surf = gsdl2.Surface(overlay_rect.size)
        gsdl2.draw.rect(surf, gsdl2.Color('yellow'), overlay_rect)
        surf.set_colorkey((0, 0, 0))
        overlay = gsdl2.texture.Texture(gsdl2.display.get_renderer(), surf)
        answer = overlay, overlay_rect
        rect_overlays[obj.radius] = answer
    return answer
rect_overlays = {}


def w(s):
    return
    stdout = sys.stdout
    stdout.write(s)
    stdout.flush()


class Game(context.Context):
    def __init__(self):
        context.Context.__init__(self, gsdl2.time.FixedDriver(self.update, settings.time_step))
        self.clock.new_schedule(self.set_caption, 0.1)
        self.clock.new_schedule(self.update_hud, 1.0)
        self.render_sched = self.clock.new_schedule(self.draw, 1.0 / settings.max_fps, keep_history=True)
        self.renderer = gsdl2.display.get_renderer()
        self.renderer.draw_color = 0, 0, 0
        gsdl2.music.set_endevent(code=settings.song_finished)

        # Space
        settings.space = spatialhash.SpatialHash(settings.screen_rect.inflate(100, 100), 64)

        # Player
        settings.player_spawn = settings.screen_rect.w // 2, settings.screen_rect.h * 0.9
        settings.player_ship = player.Ship()
        settings.space.add(settings.player_ship)

        # Sky
        settings.sky = sky.Sky()

        # Fleet HQ
        settings.enemy_fleet = enemyfleet.EnemyFleet()

        # State machine
        self.current_state = _StateBase
        self.start_state = _StateIntro

        self.paused = False
        self.mouse_pos = gsdl2.mouse.get_pos()
        self.ups = collections.deque(maxlen=int(1.0 / settings.time_step))

    def set_caption(self, dt):
        if settings.player_ship is None:
            return
        if settings.enemy_fleet is None:
            return
        try:
            gsdl2.display.set_caption('{} fps | Pause {} | Grab {} | Now playing: {}'.format(
                int(self.render_sched.per_second()), self.paused, settings.ARG_GRAB, settings.music_now_playing))
        except ZeroDivisionError:
            pass
        if not gsdl2.music.get_busy():
            if settings.play_music:
                settings.music_play()

    def update_hud(self, dt):
        hit = settings.font_cache.hit_ratio()
        num = settings.font_cache.count()
        size = int(settings.font_cache.size() / 2 ** 20)
        old_purged = settings.font_cache.old_purged
        mem_purged = settings.font_cache.mem_purged
        font1 = settings.font_ui_tiny
        color1 = settings.color_ui_tiny
        font2 = settings.font_ui_small
        color2 = settings.color_ui_small
        max_on_patrol = settings.enemy_fleet.max_on_patrol[enemy.Kamikaze]
        n_on_patrol = len(settings.n_on_patrol[enemy.Kamikaze])
        hud_info = (
            ('tex_wave', 'Wave {}'.format(settings.stage + 1), font2, color2, 255),
            ('tex_score', 'Credits {:,d}'.format(settings.score), font2, color2, 255),
        )
        if settings.internals_hud:
            hud_info += (
                ('tex_fps', '{} fps'.format(int(self.render_sched.per_second())), font2, color2, 255),
                ('tex_ups', '{} ups'.format(len(self.ups)), font1, color1, 128),
                ('tex_objs', '{} objs'.format(len(settings.space)), font1, color1, 128),
                ('tex_texts', '{} texts'.format(len(text.texts)), font1, color1, 128),
                ('tex_pshots', '{} pshots'.format(len(player.laser_shots)), font1, color1, 128),
                ('tex_eshots', '{} eshots'.format(len(enemy.laser_shots)), font1, color1, 128),
                ('tex_pexp', '{} pexplode'.format(len(player.explosions)), font1, color1, 128),
                ('tex_eexp', '{} eexplode'.format(len(enemy.explosions)), font1, color1, 128),
                ('tex_cachehit', 'Cache hit={}%'.format(hit), font1, color1, 128),
                ('tex_cachenum', 'Cache num={}'.format(num), font1, color1, 128),
                ('tex_cachesize', 'Cache size={} MB'.format(size), font1, color1, 128),
                ('tex_cachepurg', 'Purge old={} mem={}'.format(old_purged, mem_purged), font1, color1, 128),
            )
        y = 10
        for info in hud_info:
            attr, new_text, font, color, alpha = info
            old_sprite = getattr(self, attr, None)
            new_sprite = self.update_text(old_sprite, new_text, (10, y), font, color, alpha)
            setattr(self, attr, new_sprite)
            y += new_sprite.rect.h + 3

    def update_text(self, stale_sprite, new_text, pos, font, color, alpha):
        if stale_sprite:
            stale_sprite.remove()
        tex = text.TextSprite(new_text, font, color, pos, anchor='topleft', layer=settings.layer_ui_text, alpha=alpha)
        settings.space.add(tex)
        return tex

    def enter(self):
        gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(False)

    def resume(self):
        gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(False)

    def suspend(self):
        gsdl2.event.set_grab(False, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(True)

    def exit(self):
        gsdl2.event.set_grab(False, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(True)
        if settings.music_now_playing != settings.music_info[0]['name']:
            gsdl2.music.fadeout(1000)
        gsdl2.display.set_caption(settings.game_title)

    def set_state(self, new_state):
        if self.current_state:
            self.current_state.exit(self)
        new_state.enter(self)
        self.current_state = new_state

    def update(self, dt):
        if self.paused:
            self.update_paused(dt)
        else:
            settings.sky.update(dt)
            self.update_state(dt)
        self.ups.append(1)

    def update_state(self, dt):
        state = self.current_state.think(self, dt)
        if state is not None:
            self.set_state(state)
        tweener.update(dt)

    def update_paused(self, dt):
        self.toss_events()

    def update_collisions(self):
        for obj, others in settings.space.collidealldict().iteritems():
            obj.hit(others)

    def update_wave(self):
        if settings.enemy_fleet.wave_count() == 0:
            return _StateEndWave
        if not settings.player_ship.alive:
            settings.space.remove(settings.player_ship)
            return _StateRespawn

    def draw(self, sched):
        interp = sched.interp
        renderer = self.renderer
        renderer.clear()
        settings.sky.draw()
        for obj in sorted(settings.space, key=get_layer):
            obj.draw(renderer, interp)
            if settings.render_rect_overlay:
                overlay, overlay_rect = get_rect_overlay(obj)
                renderer.copy(overlay, overlay.get_rect(center=obj.pos), overlay_rect)
            if settings.render_circle_overlay:
                overlay, overlay_rect = get_circle_overlay(obj)
                renderer.copy(overlay, overlay.get_rect(center=obj.pos), overlay_rect)
        renderer.present()

    def toss_events(self):
        for e in gsdl2.event.get():
            if e.type == gsdl2.QUIT:
                self.on_quit()
            elif e.type == gsdl2.KEYDOWN:
                if e.scancode == S_ESCAPE:
                    self.on_quit()
                elif e.scancode == S_G:
                    settings.ARG_GRAB = not settings.ARG_GRAB
                    gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
                elif e.scancode == S_P:
                    self.resume()
                    self.paused = False
            elif e.type == gsdl2.MOUSEMOTION:
                if not self.paused:
                    self.on_mouse_motion(e)

    def on_mouse_motion(self, e):
        if settings.player_ship and settings.player_ship.alive:
            # settings.player_ship.mouse_motion(e.pos)
            self.mouse_pos = e.pos

    def on_mouse_button_down(self, e):
        if e.button in (1, 3):
            if settings.player_ship:
                settings.player_ship.fire(e.button)

    def on_key_down(self, e):
        if e.scancode == S_ESCAPE:
            self.on_quit()
        elif e.scancode == S_G:
            settings.ARG_GRAB = not settings.ARG_GRAB
            gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
        elif e.scancode == S_P:
            self.suspend()
            self.paused = True
        elif e.scancode == S_K:
            for ship in tuple(settings.enemy_fleet.ships_deployed):
                ship.kill(hit=True)
        elif e.scancode == S_SPACE:
            settings.sky._make_planet()

    def on_user_event(self, e):
        if e.code == settings.song_finished:
            if not gsdl2.music.get_busy():
                if settings.play_music:
                    settings.music_play()

    def on_quit(self):
        self.pop()
