__all__ = ['Texture']


import logging

import cffi

from .sdllibs import sdl_lib
from .sdlffi import sdl_ffi
from .rect import Rect
from .display import get_renderer
from . import sdlpixels


log = logging.getLogger(__name__)


class _Texture(object):
    def __init__(self, sdl_texture):
        self.__sdl_texture = sdl_texture
        format, access, w, h = self.query()
        self.__size = w, h

    def query(self):
        format = sdl_ffi.new('Uint32 *')
        access = sdl_ffi.new('int *')
        w = sdl_ffi.new('int *')
        h = sdl_ffi.new('int *')
        sdl_lib.SDL_QueryTexture(self.sdl_texture, format, access, w, h)
        # return int(format[0]), int(access[0]), int(w[0]), int(h[0])
        return format[0], access[0], w[0], h[0]

    def get_size(self):
        return self.__size
    size = property(get_size)

    def get_rect(self, **kwargs):
        w, h = self.__size
        r = Rect(0, 0, w, h)
        for k, v in kwargs.items():
            setattr(r, k, v)
        return r

    def get_blendmode(self):
        cdata = sdl_ffi.new('SDL_BlendMode *')
        sdl_lib.SDL_GetTextureBlendMode(self.sdl_texture, cdata)
        value = int(cdata[0])
        return value
    def set_blendmode(self, mode):
        sdl_lib.SDL_SetTextureBlendMode(self.sdl_texture, mode)
    blendmode = property(get_blendmode, set_blendmode)

    def get_alpha(self):
        cdata = sdl_ffi.new('Uint8 *')
        sdl_lib.SDL_GetTextureAlphaMod(self.sdl_texture, cdata)
        return cdata[0]
    def set_alpha(self, alpha):
        sdl_lib.SDL_SetTextureAlphaMod(self.sdl_texture, int(alpha))
    alpha = property(get_alpha, set_alpha)

    def __getsdltexture(self):
        return self.__sdl_texture
    sdl_texture = property(__getsdltexture)

    def __del__(self):
        # TODO: unreliable
        if self.__sdl_texture:
            try:
                garbage = self.__sdl_texture
                self.__sdl_texture = None
                sdl_lib.SDL_DestroyTexture(garbage)
            except Exception as e:
                pass


class Texture(_Texture):
    def __init__(self, size, renderer=None):
        # print('texture from scratch')
        # print('SDL_TEXTUREACCESS_TARGET', sdl_lib.SDL_TEXTUREACCESS_TARGET)
        if renderer is None:
            renderer = get_renderer()
        sdl_texture = sdl_lib.SDL_CreateTexture(
            renderer.sdl_renderer,
            sdlpixels.SDL_PIXELFORMAT_ARGB8888,
            sdl_lib.SDL_TEXTUREACCESS_TARGET, *size)
        _Texture.__init__(self, sdl_texture)


class TextureFromSurface(_Texture):
    def __init__(self, surface, renderer=None):
        # print('texture from surface')
        if renderer is None:
            renderer = get_renderer()
        sdl_texture = sdl_lib.SDL_CreateTextureFromSurface(renderer.sdl_renderer, surface.sdl_surface)
        _Texture.__init__(self, sdl_texture)


class TextureFromTexture(_Texture):
    def __init__(self, texture, renderer=None):
        # print('texture from texture')
        if not isinstance(texture, _Texture):
            print(0)
            texture = _Texture(texture)
            print(1)
        if renderer is None:
            renderer = get_renderer()
        print(2)
        # orig_target = renderer.get_target()
        print(3)
        orig_blendmode = renderer.get_blendmode()
        print(4)
        renderer.set_blendmode(sdl_lib.SDL_BLENDMODE_NONE)
        texture.set_blendmode(sdl_lib.SDL_BLENDMODE_NONE)
        rect = texture.get_rect()
        new_texture = Texture(rect.size, renderer)
        new_texture.set_blendmode(sdl_lib.SDL_BLENDMODE_NONE)
        renderer.set_target(new_texture)
        renderer.copy(texture, rect, rect)
        new_texture.set_blendmode(sdl_lib.SDL_BLENDMODE_BLEND)
        print(5)
        # if orig_target:
        #     renderer.set_target(orig_target)
        # else:
        #     renderer.set_target()
        renderer.set_blendmode(orig_blendmode)
        _Texture.__init__(self, new_texture.sdl_texture)
