from collections import deque
import os
import random

import gsdl2
import settings
from gamesound import jukebox
from geometry import angle_of, distance, point_on_circumference, circle_collided_other
import sprite
import text
import textureutil
import tweening
from vec2d import Vec2d


shooting_fore = 1   # mouse button
shooting_aft = 3    # mouse button

tweener = tweening.Tweener()
laser_shots = []
explosions = []


def update(dt):
    tweener.update(dt)
    if settings.player_ship:
        settings.player_ship.update(dt)
    for shot in laser_shots:
        shot.update(dt)
    for exp in explosions:
        exp.update(dt)


def clear():
    settings.player_ship = None
    del laser_shots[:]
    del explosions[:]
    tweener.clear()


def iround(n):
    return int(round(n))


class LaserBlast(sprite.TextureSprite):
    _texture = None
    layer = settings.layer_lasers
    collided = staticmethod(circle_collided_other)

    def __init__(self, ship, which):
        self.ship = ship
        self.speed = settings.laser_shot_speed
        if LaserBlast._texture is None:
            LaserBlast._texture = gsdl2.texture.Texture(
                gsdl2.display.get_renderer(), gsdl2.image.load(settings.img_laser_blue))
        if which == shooting_fore:
            angle = ship.angle % 360
        else:
            angle = ship.angle - 180 % 360
        sprite.TextureSprite.__init__(self, LaserBlast._texture, ship.pos, angle)
        self.alive = True

    def power_up(self, others):
        for other in others:
            other.hit(self)

    def update(self, dt):
        self.update_pos(point_on_circumference(self.pos, self.speed * dt, self.angle))
        if not settings.screen_rect.collidepoint(self.rect.center):
            self.kill()
        else:
            settings.space.add(self)

    def hit(self, others):
        for other in others:
            if isinstance(other, settings.EnemyShipTypes):
                # print('player.LaserBlast: hit', other)
                other.kill(True)
                self.kill(True)

    def kill(self, hit=False):
        if self.alive:
            self.alive = False
            laser_shots.remove(self)
            settings.space.remove(self)


class LaserCannon(object):
    def __init__(self, ship):
        self.ship = ship
        # fire rate throttles
        self.fire_rate = settings.player_fire_rate
        self.input_rate = settings.player_input_rate
        self.input_ready = True
        # mouse input queue
        self.queue = []
        self.queue_depth = settings.mouse_queue_depth
        # shots on screen
        self.max_shots = settings.player_max_shots

    def update(self, dt):
        self.input_rate -= dt
        if self.input_rate <= 0:
            self.input_ready = True
        if self.fire_rate > 0:
            self.fire_rate -= dt
        if len(self.queue) > 0:
            if self.fire_rate <= 0 and len(laser_shots) < self.max_shots:
                self._fire()

    def fire(self, which):
        assert which in (shooting_fore, shooting_aft)
        if self.input_ready and len(self.queue) < self.queue_depth and len(laser_shots) < self.max_shots:
            self.queue.append(which)
            self.input_ready = False
            self.input_rate = settings.player_input_rate

    def _fire(self):
        jukebox['player_laser'].play()
        blast = LaserBlast(self.ship, self.queue.pop(0))
        laser_shots.append(blast)
        self.fire_rate = settings.player_fire_rate
        settings.space.add(blast)


class ShipExplodes(sprite.TextureSprite):
    layer = settings.layer_explosions
    textures = []

    def __init__(self, pos):
        jukebox['player_hit'].play()
        if not ShipExplodes.textures:
            sheet = gsdl2.image.load(settings.img_player_explode)
            sheet.set_colorkey((0, 0, 0))
            sheet = gsdl2.Texture(gsdl2.display.get_renderer(), sheet)
            ShipExplodes.textures[:] = textureutil.copy_strip(sheet, gsdl2.Rect(1, 1, 96, 96), numx=4, numy=2)
        sprite.TextureSprite.__init__(self, self.textures[0], pos, angle=random.random() * 360.0, alpha=32*7)

        self.tween_val = 0
        self.tween = tweener.create_tween(self, 'tween_val', 0, 8, 1.0, cb_end=self.remove)

        explosions.append(self)
        settings.space.add(self)

    def hit(self, other):
        pass

    def update(self, dt):
        try:
            self.texture = self.textures[int(self.tween_val)]
        except IndexError:
            pass

    def remove(self, *args):
        try:
            explosions.remove(self)
        except ValueError:
            pass
        settings.space.remove(self)


def print_surf(surf):
    s = surf.sdl_surface
    f = s.format
    flags = f.Rmask, f.Gmask, f.Bmask, f.Amask
    shifts = f.Rshift, f.Gshift, f.Bshift, f.Ashift
    print(flags, shifts)


class Ship(sprite.TextureSprite):
    _texture = None
    layer = settings.layer_player
    collided = staticmethod(circle_collided_other)

    def __init__(self):
        if Ship._texture is None:
            sheet = gsdl2.image.load_texture(settings.img_player)
            w, h = 256, 256
            Ship._texture = textureutil.copy(sheet, gsdl2.Rect(0, h, w, h).scale(0.5, 0.5), scale=0.48)
        game_rect = settings.screen_rect
        w, h = Ship._texture.size
        sprite.TextureSprite.__init__(self, Ship._texture, (game_rect.centerx, game_rect.height - h))

        self.weapon = LaserCannon(self)

        # nose is the ship's nose. mouse pos will be maintained at this position to calculate move_to and angle.
        # turn_to_angle is the target angle for smoothness of turning visual.
        # See mouse_motion() and turn().
        self.nose = point_on_circumference(self.pos, self.radius, self.angle)
        self.turn_to_angle = self.angle
        gsdl2.mouse.set_pos(*self.nose)

        self.alive = True

    def update(self, dt):
        if not self.alive:
            return
        self.weapon.update(dt)
        if self.pos != self.move_to:
            speed = settings.player_speed * dt
            dist = distance(self.pos, self.move_to)
            if dist <= speed:
                angle = angle_of(self.pos, self.move_to)
                step = point_on_circumference(self.pos, speed * 0.2, angle)
                if step >= dist:
                    step = self.move_to
            else:
                angle = angle_of(self.pos, self.move_to)
                step = point_on_circumference(self.pos, speed, angle)
            self.update_pos(step)
            settings.space.add(self)
            self.update_nose()
        else:
            self.update_pos(self.pos)
            self.update_nose()
        self.turn(dt)

    def update_nose(self):
        self.nose = point_on_circumference(self.pos, self.radius, self.turn_to_angle)
        gsdl2.mouse.set_pos(*self.nose)

    def mouse_motion(self, pos):
        angle = angle_of(self.pos, pos)
        radius = self.radius
        self.nose = point_on_circumference(self.pos, radius, angle)
        self.turn_to_angle = angle
        dist = distance(self.pos, pos)
        if dist > radius:
            self.move_to = point_on_circumference(self.pos, dist - radius, angle)
        gsdl2.mouse.set_pos(*self.nose)

    def turn(self, dt):
        self.update_nose()
        if self.angle != self.turn_to_angle:
            diff = self.turn_to_angle - self.angle
            if diff < -180:
                diff += 360
            elif diff >= 180:
                diff -= 360
            turn_rate = settings.player_turn_rate * dt
            if diff <= -turn_rate:
                self.angle -= turn_rate
            elif diff >= turn_rate:
                self.angle += turn_rate
            else:
                self.angle = self.turn_to_angle
            self.angle %= 360.0

    def fire(self, which):
        self.weapon.fire(which)

    def hit(self, others):
        for other in others:
            if isinstance(other, settings.EnemyTypes):
                # print('player.Ship: hit', other)
                other.kill(True)
                self.kill(True)

    def kill(self, hit=False):
        if self.alive:
            self.alive = False
            settings.space.remove(self)
            ShipExplodes(self.pos)
            # Penalty multiplies every 5 levels
            penalty = settings.player_penalty
            penalty += (settings.stage + 1) // 5 * penalty
            settings.score += penalty
            text.make_floating_score(str(penalty), self.rect.midtop)

    # def power_up(self, others):
    #     for other in others:
    #         other.hit(self)

    def reset(self):
        self.update_pos(settings.player_spawn, interp=False)
        self.angle = 0
        self.weapon = LaserCannon(self)
        for shot in list(laser_shots):
            shot.kill()


# Initialize classifiers in settings module
settings.FriendlyShipTypes = (
    Ship,
)
settings.FriendlyFireTypes = (
    LaserBlast,
)
settings.FriendlyTypes = settings.FriendlyShipTypes + settings.FriendlyFireTypes
