#   Trolls Outta Luckland v1.1.0
#   Copyright 2009-2010, Gummbum
#
#   This file is part of Trolls Outta Luckland.
#
#   Trolls Outta Luckland is free software: you can redistribute it
#   and/or modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation, either version 3 of
#   the License, or (at your option) any later version.
#
#   Trolls Outta Luckland is distributed in the hope that it will be
#   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Trolls Outta Luckland.  If not, see
#   <http://www.gnu.org/licenses/>.

__version__ = '1.1.0'

import collections
from os.path import join as join_path
import random
from random import randrange

import gsdl2

import settings
import sprite
from vec2d import Vec2d


# These are NASA and art pics in the public domain.
_PLANET_PICS = [
    'PIA01666.png',
    'PIA01667.png',
    'PIA10102.png',
    'PIA10103.png',
    'PIA11667.png',
    'Saturn_from_Cassini_Orbiter_(2004-10-06).png',
]
_GALAXY_PICS = [
    '600px-Andromeda_galaxy.png',
    '800px-Hubble2005-01-barred-spiral-galaxy-NGC1300.png',
    '800px-M104_ngc4594_sombrero_galaxy_hi-res.png',
    '800px-Messier_81_HST.png',
    '8d8e6_spaceart.png',
    'HubbleInteractingGalaxyNGC60506_3.png',
    'post_1611095_1251219044_med.png',
    'spinning-galaxy.png',
]


def iround(n):
    return int(round(n))


class Sky(object):
    def __init__(self, max_stars=30, max_planets=1, max_galaxies=3):
        self.sprites = collections.defaultdict(list)
        self.max_stars = max_stars
        self.max_planets = max_planets
        self.max_galaxies = max_galaxies
        self.onscreen = []
        self.reset()

    def _stars(self):
        return self.get_sprites_from_layer(0)

    def _galaxies(self):
        return self.get_sprites_from_layer(1)

    def _planets(self):
        return self.get_sprites_from_layer(2)

    def get_sprites_from_layer(self, layer):
        if layer not in self.sprites:
            return []
        return list(self.sprites[layer])

    def add(self, sprite):
        self.sprites[sprite.layer].append(sprite)

    def _make_star(self, randy=False):
        self.add(Star(randy=True))

    def _make_galaxy(self, randy=False):
        random.shuffle(_GALAXY_PICS)
        for name in _GALAXY_PICS:
            if name not in self.onscreen:
                self.add(Galaxy(name, randy=True))
                self.onscreen.append(name)
                break

    def _make_planet(self):
        random.shuffle(_PLANET_PICS)
        for name in _PLANET_PICS:
            if name not in self.onscreen:
                # print('Planet: {}'.format(name))
                self.add(Planet(name))
                self.onscreen.append(name)
                break

    def reset(self, debug=False):
        self.sprites.clear()
        del self.onscreen[:]
        for i in range(self.max_stars):
            self._make_star(randy=True)
        if debug:
            for i in range(len(self._planets()), self.max_planets):
                self._make_planet()
        for i in range(self.max_galaxies):
            self._make_galaxy(randy=True)

    def update(self, dt):
        for i in range(len(self._stars()), self.max_stars):
            if randrange(int(2 / dt) + 1) == 0:
                self._make_star()
        # 60 secs
        for i in range(len(self._planets()), self.max_planets):
            if randrange(int(60 / dt) + 1) == 0:
                self._make_planet()
        # 60 seconds
        for i in range(len(self._galaxies()), self.max_galaxies):
            if randrange(int(60 * dt) + 1) == 0:
                self._make_galaxy()
        for layer in sorted(self.sprites):
            for obj in self.sprites[layer]:
                obj.update(dt)

    def draw(self):
        renderer = gsdl2.display.get_renderer()
        for layer in sorted(self.sprites):
            for obj in self.sprites[layer]:
                obj.draw(renderer)


class Star(sprite.TextureSprite):
    texture = None

    def __init__(self, *args, **kwargs):
        randy = kwargs.get('randy', False)
        if Star.texture is None:
            Star.texture = gsdl2.image.load_texture(settings.img_star)
        screen_rect = settings.screen_rect
        x = randrange(screen_rect.right)
        y = float(randrange(screen_rect.y, screen_rect.bottom)) if randy else -5.0
        sprite.TextureSprite.__init__(self, Star.texture, (x, y))
        self.speed = 0.33
        self.alive = True

    def update(self, dt):
        self.update_pos((self.pos.x, self.pos.y + self.speed * dt))
        if self.pos.y >= settings.screen_rect.bottom + 10:
            self.kill()
        # if (self.alpha_step < 0 and self.alpha <= self.alpha_target) or \
        # (self.alpha_step >= 0 and self.alpha >= self.alpha_target):
        #     self.alpha_target = rand(*self.alpha_range)
        #     if self.alpha < self.alpha_target:
        #         self.alpha_step = rand(2,8)
        #     else:
        #         self.alpha_step = -rand(2,8)
        # self.alpha += self.alpha_step
        # self.image.set_alpha(self.alpha)

    def kill(self, *args):
        if not self.alive:
            return
        self.alive = False
        if self in settings.sky.onscreen:
            settings.sky.onscreen.remove(self)
        if self in settings.sky.sprites:
            settings.sky.sprites.remove(self)


class Body(sprite.TextureSprite):
    def __init__(self, *args, **kwargs):
        """kwargs:
            image=filename
            colorkey=color, -1, or None; default -1
            alpha=0 to 255; default 255
            scale=0.0 to 1.0; default 1.0
            rotate=0 to 360; default 0
            pos=(x,y)
            speed=(dx,xy)
        """
        name = kwargs.get('image')
        color = kwargs.get('colorkey', -1)
        image = gsdl2.image.load(join_path('data/image', name))
        screen_rect = settings.screen_rect
        x, y = screen_rect.topleft
        r, b = screen_rect.bottomright
        try:
            if color == -1:
                color = image.get_at((0, 0))
        except:
            color = gsdl2.Color('black')
        image.set_colorkey(color)
        texture = gsdl2.texture.Texture(gsdl2.display.get_renderer(), image)
        sprite.TextureSprite.__init__(
            self, texture,
            kwargs.get('pos', (randrange(x, r), randrange(y, b))),
            kwargs.get('rotate', 0.0),
            kwargs.get('scale', 1.0),
            kwargs.get('alpha', 255))
        self.speed = Vec2d(kwargs.get('speed'))
        self.alive = True

    def update(self, dt):
        speed = self.speed * dt
        self.update_pos(self.pos + speed)
        if not self.rect.colliderect(settings.screen_rect):
            self.kill()

    def kill(self, *args):
        if not self.alive:
            return
        self.alive = False
        if self in settings.sky.onscreen:
            settings.sky.onscreen.remove(self)
        if self in settings.sky.sprites:
            settings.sky.sprites.remove(self)


class Planet(Body):
    def __init__(self, image_filename):
        scale = randrange(50, 100) / 100.0
        rotate = randrange(360)
        x = randrange(100, settings.screen_rect.right - 100)
        y = -130
        dy = 0.5 / settings.time_step * scale**2
        dx = 0.25 / settings.time_step * scale**2
        dx *= random.choice((-1, 1))
        Body.__init__(self, image=image_filename, rotate=rotate, scale=scale,
                      pos=(x, y), speed=(dx, dy))


class Galaxy(Body):
    def __init__(self, image_filename, randy=False):
        self.name = image_filename
        scale = randrange(50, 100) / 100.0
        x = randrange(64, settings.screen_rect.right - 128)
        if randy:
            y = randrange(64, settings.screen_rect.bottom - 128)
        else:
            y = -64
        speed = randrange(2, 5) / 10.0
        dx = 0
        dy = settings.time_step * float(speed + 0.1) / 240
        Body.__init__(self, image=image_filename, alpha=80,
                      rotate=randrange(0, 359), scale=scale, pos=(x, y), speed=(dx, dy))
