import gsdl2

import tweening
from vec2d import Vec2d


tweener = tweening.Tweener()


def update(dt):
    tweener.update(dt)


def clear():
    tweener.clear()


class TextureSprite(object):
    layer = 0

    def __init__(self, texture, pos, angle=0.0, scale=1.0, alpha=255, radius=-1, flip=False, anchor='center'):
        self._texture = texture
        self.rect = texture.get_rect(**{anchor: pos})
        self.src_rect = texture.get_rect()
        self._pos = Vec2d(pos)
        self._old_pos = Vec2d(pos)
        self._move_to = Vec2d(pos)
        self.angle = angle
        self.scale = scale
        self.alpha = alpha
        self.anchor = anchor
        self.radius = self.rect.w // 2 if radius == -1 else radius
        self.flip = flip

        self.tweens = {}
        self.driftx = 0.0
        self.drifty = 0.0

    def _get_origin(self):
        return self.pos
    origin = property(_get_origin)

    def _get_pos(self):
        return self._pos
    def _set_pos(self, val):
        self._pos[:] = val
    pos = property(_get_pos, _set_pos)

    def _get_move_to(self):
        return self._move_to
    def _set_move_to(self, val):
        self._move_to[:] = val
    move_to = property(_get_move_to, _set_move_to)

    def _get_old_pos(self):
        return self._old_pos
    def _set_old_pos(self, val):
        self._old_pos[:] = val
    old_pos = property(_get_old_pos, _set_old_pos)

    def _get_texture(self):
        self._texture.alpha = getattr(self, 'alpha', 255)
        return self._texture
    def _set_texture(self, val):
        self._texture = val
    texture = property(_get_texture, _set_texture)

    def update_pos(self, val, interp=True):
        self.old_pos = self.pos if interp else val
        self.pos = self.move_to = val
        setattr(self.rect, self.anchor, val)

    def hit(self, *args):
        """override"""
        pass

    def kill(self, *args):
        """override"""
        pass

    def update(self, dt):
        """override"""
        pass

    def draw(self, renderer, interp=1.0):
        scale = self.scale
        rect = self.rect if scale == 1.0 else self.rect.scale(scale, scale)
        if 0.0 <= interp < 1.0:
            # interpolate position
            fm = self.old_pos
            to = self.pos
            if fm != to:
                dx = (to.x - fm.x) * interp
                dy = (to.y - fm.y) * interp
                rect = rect.move(dx, dy)
        src_rect = self.src_rect
        renderer.copy_ex(self.texture, rect, src_rect, self.angle, getattr(src_rect, self.anchor), self.flip)
        # Debug: draw bounding rect and print info
        # renderer.set_draw_color(gsdl2.Color('white'))
        # gsdl2.sdl_lib.SDL_RenderDrawRect(renderer.sdl_renderer, rect.sdl_rect)
        # if self.__class__.__name__ == 'Ship':
        #     r = rect.sdl_rect
        #     print('sprite: {} {}'.format(rect, (r.x, r.y, r.w, r.h)))
        #     print('sprite: {} {}'.format(self.scale, self.texture.size))
        #     quit()
        # renderer.set_draw_color(gsdl2.Color(0, 0, 0))

    def fade_effect(self, start_alpha=255, end_alpha=128, duration=1.0, tween_function=tweening.LINEAR, cb_end=None):
        self.misc_effect('alpha', start_alpha, end_alpha, duration, tween_function, cb_end)
        return self

    def spin_effect(self, start_angle=0.0, end_angle=360.0, duration=1.0, tween_function=tweening.LINEAR, cb_end=None):
        self.misc_effect('angle', start_angle, end_angle, duration, tween_function, cb_end)
        return self

    def scale_effect(self, start_scale=1.0, end_scale=2.0, duration=1.0, tween_function=tweening.LINEAR, cb_end=None):
        self.misc_effect('scale', start_scale, end_scale, duration, tween_function, cb_end)
        return self

    def driftx_effect(self, start_x=0.0, end_x=0.0, duration=1.0, tween_function=tweening.LINEAR, cb_end=None):
        self.misc_effect('driftx', start_x, end_x, duration, tween_function, cb_end)
        return self

    def drifty_effect(self, start_y=0.0, end_y=0.0, duration=1.0, tween_function=tweening.LINEAR, cb_end=None):
        self.misc_effect('drifty', start_y, end_y, duration, tween_function, cb_end)
        return self

    def misc_effect(self, attr_name, start_val, end_val, duration, tween_function=tweening.LINEAR, cb_end=None):
        """

        :param attr_name: str; the attribute created on this object to store the computed value
        :param start_val: int or float; starting value for the tween
        :param end_val: int or float; ending value for the tween
        :param duration: int or float; duration of the tween in seconds
        :param tween_function: see tweening module for available functions
        :param cb_end: list; callback args; see TextureSprite._cb_end() for handling the callback args
        :return:
        """
        if cb_end is None:
            cb_end = self._cb_end
        self.tweens[attr_name] = tweener.create_tween(
            self, attr_name, start_val, end_val - start_val, duration,
            tween_function, cb_end=cb_end, cb_args=[attr_name])
        return self

    def end_fade(self):
        self.end_misc('alpha')

    def end_spin(self):
        self.end_misc('angle')

    def end_scale(self):
        self.end_misc('scale')

    def end_driftx(self):
        self.end_misc('driftx')
        self.driftx = 0.0

    def end_drifty(self):
        self.end_misc('drifty')
        self.drifty = 0.0

    def end_misc(self, attr_name):
        tweener.remove_tween(self, attr_name)
        if attr_name in self.tweens:
            del self.tweens[attr_name]

    def _cb_end(self, *args):
        # args: see tweening.Tweener.remove_tween()
        tween, obj, attr_name, begin, change, dur, fn, params, cb, cb_args = args
        if attr_name in self.tweens:
            del self.tweens[attr_name]
